# -*- coding: utf-8 -*-
"""
    drupan.engine

    The engine is part of drupan that holds all plugins and subsystems
    together and runs the actual site creation process.
"""

import os
import sys

from .site import Site
from .config import Config
from .template import Render
from .serve import server


class Engine(object):
    """Drupan engine doing all the work"""
    def __init__(self):
        self.site = Site()
        self.config = Config()
        self.reader = None
        self.writer = None
        self.plugins = list()
        self.renderer = None
        self.deployment = None

    def prepare_engine(self):
        """get all subsystems and plugins setup"""
        self.add_external_plugins()

        if self.config.reader:
            imported = self._load_module(self.config.reader, "inout", "Reader")
            self.reader = imported.Reader(self.site, self.config)

        if self.config.writer:
            imported = self._load_module(self.config.writer, "inout", "Writer")
            self.writer = imported.Writer(self.site, self.config)

        for name in self.config.plugins:
            imported = self._load_module(name, "plugins", "Plugin")
            plugin = imported.Plugin(self.site, self.config)
            self.plugins.append(plugin)

        self.renderer = Render(self.site, self.config)

        if self.config.deployment:
            imported = self._load_module(
                self.config.deployment,
                "deployment",
                "Deploy",
            )
            self.deployment = imported.Deploy(self.site, self.config)

    @staticmethod
    def _load_module(name, base_name, kind):
        """Load a drupan module and return it. First try to load a module from
        path with the format `drupan-$pluginName` before trying to import a
        plugin from the drupan distribution.

        If both imports fail let drupan run into an ImportError exception since
        something is clearly broken.

        :param name: name of the module to load
        :param base_name: base path for drupan standard module
        :param kind: class to import
        :returns: imported class from `name`
        """
        try:
            plugin_name = "drupan-{0}".format(name)
            return __import__(plugin_name, fromlist=[kind])
        except ImportError:
            plugin_name = "drupan.{0}.{1}".format(base_name, name)
            return __import__(plugin_name, fromlist=[kind])

    def add_external_plugins(self):
        """Add external plugin path to Python path."""
        if not self.config.external_plugins:
            return

        plugin_path = self.config.external_plugins

        if not os.path.exists(plugin_path):
            raise Exception("External plugin path does not exist.")

        sys.path.append(plugin_path)

    def run(self):
        """run the site generation process"""
        self.reader.run()
        for plugin in self.plugins:
            plugin.run()
        self.renderer.run()
        self.writer.run()

    def serve(self):
        """serve the generated site"""
        server(self.config.get_option("writer", "directory"))

    def deploy(self):
        """deploy the generated site"""
        if not self.deployment:
            return

        self.deployment.run()



# -*- coding: utf-8 -*-


class FileWrapper(object):
    """FileWrapper provides an interface to move files through drupan.

    Every reader plugin should wrap files, like images, in an instance  of a
    class which inherits FileWrapper which provides memory efficient access to
    the files content.

    Every plugin that needs to know if it is dealing with a file should check
    if the variable is an instance of FileWrapper.
    """
    def read(self):
        """:returns: content of the file"""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    drupan.deployment.s3cf

    Deploying the website to AWS S3 and CloudFront. This plugin will send
    and invalidation request to CloudFront once the deployment is complete,
    invalidating *all* files that changed.

    Using CloudFront invalidation requests can result in high bills. Please
    consider this before using this plugin. If this is a deal breaker use
    s3sub with a 5 minute cache policy.
"""

from hashlib import md5
from mimetypes import MimeTypes

import boto
from boto.s3.connection import S3Connection, OrdinaryCallingFormat
from boto.s3.key import Key

from drupan.file_wrapper import FileWrapper


S3_HOST_ERROR = "buckets name contains a full stop.\n\n"
S3_HOST_ERROR += "Please set the `s3_host` configuration key to your S3 host."


class S3HostMissingException(Exception):
    pass


class Deploy(object):
    def __init__(self, site, config):
        """
        Arguments:
            site: generated site
            config: config for this site
        """
        self.site = site
        self.config = config

        self.bucket_name = config.get_option("s3cf", "bucket")
        self.aws_access_key = config.get_option("s3cf", "aws_access_key")
        self.aws_secret_key = config.get_option("s3cf", "aws_secret_key")
        self.cloudfront_id = config.get_option(
            "s3cf",
            "cloudfront_id",
            optional=True,
        )
        self.s3_host = None

        if "." in self.bucket_name:
            try:
                self.s3_host = config.get_option("s3cf", "s3_host")
            except Exception:
                raise S3HostMissingException(S3_HOST_ERROR)

        self.s3_connection = None
        self.bucket = None
        self.cf_connection = None
        self.to_invalidate = list()

        self.redirects = config.get_option("s3cf", "redirects", optional=True)

        if self.redirects is None:
            self.redirects = config.get_option(None, "redirects")

    def run(self):
        """run the deployment process"""
        self.setup()

        self.upload_entities()
        self.upload_assets()
        self.upload_templates()
        self.invalidate()
        self.redirect()

    def setup(self):
        """setup AWS connection"""
        if "." in self.bucket_name:  # work around for a boto bug
            self.s3_connection = S3Connection(
                self.aws_access_key,
                self.aws_secret_key,
                calling_format=OrdinaryCallingFormat(),
                host=self.s3_host,
            )
        else:
            self.s3_connection = S3Connection(
                self.aws_access_key,
                self.aws_secret_key,
            )
        self.bucket = self.s3_connection.get_bucket(self.bucket_name)

        if self.cloudfront_id:
            self.cf_connection = boto.connect_cloudfront(
                self.aws_access_key,
                self.aws_secret_key,
            )

    def upload_entities(self):
        """upload changed files to S3"""
        for entity in self.site.entities:
            self.upload(
                entity.file_path,
                entity.rendered.encode("utf-8"),
                entity.url,
            )
            self.upload_images(entity)

    def upload_images(self, entity):
        """upload all images for this entity

        :param entity: entity to upload images for
        """
        for name in entity.images:
            path = entity.path + "/" + name
            self.upload(path, self.site.images[name])

    def upload_assets(self):
        """upload assets stored in template directory"""
        for name, asset in self.site.assets.items():
            if name.startswith("_"):
                continue

            self.upload(name, asset)

    def upload_templates(self):
        """upload template files which are not prefixed with _"""
        for name, content in self.site.templates.items():
            if name.startswith("_"):
                continue

            self.upload(name, content, name)

    def upload(self, path, content, invalidate=None):
        """
        Set the content, mime type and ACL for a key on S3. Before setting the
        check if the object is new or changed.


        :param path: path for key
        :param content: content to set
        :param invalidate: CloudFront path to add to invalidation list. * will
                           be added to the end to make sure we invalidate the
                           URL path with a trailing slash and the html itself.
                           If None the path will be used.
        """
        changed = self.file_changed(path, content)

        if not changed:
            return

        key = Key(self.bucket)
        key.content_type = guess_mime_type(path)
        key.key = path
        if isinstance(content, FileWrapper):
            key.set_contents_from_string(content.read())
        else:
            key.set_contents_from_string(content)
        key.set_acl("public-read")

        print("uploaded: {0}".format(path))

        if invalidate is None:
            invalidate = path

        self.to_invalidate.append(invalidate + "*")

    def file_changed(self, path, content):
        """
        Check if an keys content is new or changed. If the Key is None the file
        is new, if not compare the etag (MD5) with the checksum of the files
        content.

        Arguments:
            path: key path
            content: content to compare

        Returns:
            True if the key changes or is new and needs to be uploaded and
            invalidated.
        """
        key = self.bucket.get_key(path)

        if not key:
            return True

        if key.etag is None:
            return True

        if isinstance(content, FileWrapper):
            checksum = md5(content.read()).hexdigest()
        else:
            checksum = md5(content).hexdigest()
        return key.etag.replace('"', "") != checksum

    def invalidate(self):
        """Invalidate changed entities"""
        if self.cloudfront_id is None:
            return

        self.cf_connection.create_invalidation_request(
            self.cloudfront_id,
            self.to_invalidate,
        )

        print("invalidating: {0}".format(self.to_invalidate))

    def redirect(self):
        """create a redirect"""
        if self.redirects == dict() or self.redirects is None:
            print("No redirects - skipping.")
            return

        for source, destination in self.redirects.items():
            if self.redirect_exists(source):
                continue

            key = Key(self.bucket)
            key.key = source
            key.set_redirect(destination)
            key.set_acl('public-read')

            print("redirecting: {0} -> {1}".format(source, destination))

    def redirect_exists(self, redirect):
        """
        Check if a redirect exists.

        :param redirect: redirect URL to check

        :returns: True if the redirect already exists
        """
        if self.bucket.get_key(redirect):
            return True
        return False


def guess_mime_type(name):
    """
    Guess the mime type for a filename. Default: binary/octet-stream

    Arguments:
        name: name of the file to return mime type for

    Returns:
        mime type as string
    """
    mime = MimeTypes()
    typ = mime.guess_type(name)[0]

    if typ is None:
        return "binary/octet-stream"

    return typ



# -*- coding: utf-8 -*-
"""
    drupan.io.filesystem

    Filesystem implements a filesystem reader and writer. It reads drupans
    content files from the your disk and writes the generated site back to it.
"""

import os
from io import open
import shutil
import errno
import re

import yaml

from drupan.entity import Entity
from drupan.file_wrapper import FileWrapper


class Reader(object):
    """
    Implements a filesystem reader parsing drupans standard format. It consists
    of a header with all meta data, separated by three dashes from the content.
    """
    def __init__(self, site, config):
        """
        Arguments:
            site: instance of drupan.site.Site
            config: instance of drupan.config.Config
        """
        self.site = site
        self.config = config
        self.content = config.get_option("reader", "content", optional=True)
        self.extension = config.get_option("reader", "extension")
        self.template = config.get_option("reader", "template", optional=True)

        if self.content is None:
            # Try getting the content directory via the content key. This
            # preserves 2.0 config format. Will be removed in 3.0
            self.content = config.get_option("reader", "directory")
            print("Please rename your directory key to content.")

        if self.template is None:
            # If the reader is not configured for templates try getting it
            # via the jinja key. This preserves the behavior of the 2.0 config
            # format. Will be removed in 3.0
            self.template = config.get_option("jinja", "template")
            print("Please move your template key to the reader section.")

        if not self.extension.startswith("."):
            self.extension = ".{0}".format(self.extension)

    def run(self):
        """read and parse all files"""
        self.read_content()
        self.read_template()
        self.read_images()

    def read_content(self):
        """read the content files and create entities"""
        for current in os.listdir(self.content):
            if not os.path.splitext(current)[1] == self.extension:
                continue

            fqp = os.path.join(self.content, current)

            with open(fqp, 'r', encoding='utf-8') as infile:
                self.parse_file(infile.read())

    def parse_file(self, raw):
        """
        create an entity from a read file

        Arguments:
            raw: input
        """
        (header, separator, content) = raw.partition("---")
        meta = yaml.load(header)

        entity = Entity(self.config)
        entity.meta = meta
        entity.raw = content.strip()

        self.site.entities.append(entity)

    def read_template(self):
        """read template files and populate template dict"""
        template_extensions = (".html", ".htm", ".xml", ".txt")

        for file_names, dir_name in dir_walker(self.template):
            for name in file_names:
                if name.startswith("."):
                    continue

                store = self.site.templates
                mode = "r"

                if not name.endswith(template_extensions):
                    store = self.site.assets
                    mode = "rb"

                # remove template dir part from directory name
                stripped = re.sub(self.template, "", dir_name)

                # remove leading separator if exists
                if stripped.startswith(os.path.sep):
                    stripped = stripped[1:]

                fqp = os.path.join(dir_name, name)
                key = os.path.join(stripped, name)

                if name.endswith(template_extensions):
                    with open(fqp, mode) as infile:
                        store[key] = infile.read()
                else:
                    store[key] = FileSystemFileWrapper(fqp)

    def read_images(self):
        """read images and populate images dict"""
        path = os.path.join(self.content, "images")

        for file_names, dir_name in dir_walker(path):
            for name in file_names:
                if name.startswith("."):
                    continue

                fqp = os.path.join(dir_name, name)
                self.site.images[name] = FileSystemFileWrapper(fqp)


class Writer(object):
    """
    Implements a filesystem writer, writing all entities which have a rendered
    attribute != None. The path is the URL split and appended to the base path.
    """
    def __init__(self, site, config):
        """
        Arguments:
            site: instance of drupan.site.Site
            config: instance of drupan.config.Config
        """
        self.site = site
        self.config = config

        self.base_path = config.get_option("writer", "directory")

    def run(self):
        """run the plugin"""
        self.clean_dir()
        self.write_entities()
        self.write_assets()
        self.write_templates()

    def write_entities(self):
        """write entities to files"""
        for entity in self.site.entities:
            if entity.path is None:
                continue

            path = os.path.join(self.base_path, entity.path)
            create_path(path)

            path = os.path.join(path, "index.html")
            write(entity.rendered, path)

            self.write_images(entity)

    def write_assets(self):
        """write assets to files"""
        for name, content in self.site.assets.items():
            path = os.path.join(self.base_path, os.path.dirname(name))
            create_path(path)

            path = os.path.join(self.base_path, name)
            write(content, path)

    def write_templates(self):
        """write template files which are not prefixed with _"""
        for name, content in self.site.templates.items():
            if name.startswith("_"):
                continue

            path = os.path.join(self.base_path, os.path.dirname(name))
            create_path(path)

            path = os.path.join(self.base_path, name)
            write(content, path)

    def write_images(self, entity):
        """
        Write images that are linked in this entity to the entity folder

        Arguments:
            entity: entity to copy images to
        """
        path = os.path.join(self.base_path, entity.path)

        for name in entity.images:
            image = self.site.images[name]
            output = os.path.join(path, name)
            write(image, output)

    def clean_dir(self):
        """clean the output directory"""
        try:
            shutil.rmtree(self.base_path)
        except OSError:
            pass


class FileSystemFileWrapper(FileWrapper):
    def __init__(self, fqp):
        """Wrap a local file. Store its file object in an instance variable.

        :param fqp: full path to file.
        """
        self._file_object = open(fqp, "rb")

    def read(self):
        """:returns: content of the file"""
        self._file_object.seek(0)
        return self._file_object.read()


def dir_walker(path):
    """
    Use os.walk to walk through directory structure. Exclude directories
    starting with a . and yield file names and the directory name.

    Arguments:
        path: path to walk
    """
    for dir_name, dir_names, file_names in os.walk(path):
        skip = False

        # ignore dictionaries starting with a .
        for name in dir_name.split(os.path.sep):
            if name.startswith("."):
                skip = True

        if skip is True:
            continue

        yield (file_names, dir_name)


def create_path(path):
    """
    Create all directories needed to write an entity.

    :param path: path to create
    """
    if path == "":
        return

    if os.path.exists(path):
        return

    try:
        os.makedirs(path)
    except OSError:
        if OSError.errno == errno.EEXIST:
            pass
        else:
            raise


def write(content, path):
    """
    Write an entity to disk

    Arguments:
        content: content to write
        path: path to write to
    """
    if isinstance(content, FileWrapper):
        with open(path, "wb") as output:
            output.write(content.read())
    else:
        with open(path, "w", encoding="utf-8") as output:
            output.write(content)



# -*- coding: utf-8 -*-
from html.parser import HTMLParser


class ImageParser(HTMLParser):
    """Handler based on HTMLParser for images"""

    def __init__(self):
        HTMLParser.__init__(self)
        self.images = []
        self.exclude = ("/", "//", "https://", "http://", "ftp://")

    def handle_starttag(self, tag, attrs):
        """if image tag is found add it to self.images"""
        if tag == 'img':
            img = dict(attrs)['src']

            if not img.startswith(self.exclude):
                self.images.append(img)



# -*- coding: utf-8 -*-
"""
    drupan.entity

    Each page / post / snippet / ... is defined as one entity.
"""

import re
from datetime import datetime

from drupan.imageparser import ImageParser


class Entity(object):
    """define an entity and all helper methods"""
    def __init__(self, config):
        self.config = config

        self.meta = dict()
        self.raw = None
        self.content = None

        # store results generated by the matching property
        self._slug = None
        self._url = None
        self._created = None
        self._updated = None
        self._content_type = None
        self._images = None

    def __repr__(self):
        return self.url

    @property
    def layout(self):
        """
        Returns:
            layout for this entity
        """
        return self.meta.get("layout", None)

    @property
    def url(self):
        """
        Return the URL for this entity. If there is no URL key in the meta
        dictionary config["urls"][layout] will be used.

        Returns:
            url for this entity
        """
        if self._url:
            return self._url

        if "url" in self.meta:
            layout = self.meta["url"]
        elif self.layout in self.config.url_scheme:
            layout = self.config.url_scheme[self.layout]
        else:
            return None

        for key in layout.split("/"):
            if not len(key) > 0:
                continue

            if not key.startswith("%"):
                continue

            value = self.get_url_value(key)
            layout = layout.replace(key, value)

        if len(layout) > 0 and not layout.endswith("/"):
            layout += "/"

        if len(layout) > 0 and not layout.startswith("/"):
            layout = "/" + layout

        self._url = layout
        return self._url

    @property
    def slug(self):
        """
        Returns:
            slug based on the title
        """
        if self._slug:
            return self._slug

        clean = re.sub('[^A-Za-z0-9]+', '-', self.meta.get("title", ""))

        # multiple '-' do not look nice
        clean = clean.replace('----', '-')
        clean = clean.replace('---', '-')
        clean = clean.replace('--', '-')

        # urls should not start or end with a -
        if clean.endswith("-"):
            clean = clean[:-1]

        if clean.startswith("-"):
            clean = clean[1:]

        clean = clean.lower()

        self._slug = clean
        return self._slug

    @property
    def created(self):
        """
        Returns:
            create date and time as datetime instance
        """
        if self._created:
            return self._created

        if "date" in self.meta:
            dt = self.meta["date"]
        else:
            dt = datetime.now()

        self._created = dt

        return self._created

    @property
    def updated(self):
        """
        Returns:
            update date and time as datetime instance
        """
        if self._updated:
            return self._updated

        dt = self.meta.get("updated", None)

        # update timestamp is not required - use the created timestamp
        if not dt:
            self._updated = self.created
            return self._updated

        self._updated = dt
        return self._updated

    @property
    def path(self):
        """
        Returns:
            path without leading or trailing slash
        """
        if self.url is None:
            return None

        if self.url.rsplit("/", 1)[0] != "":
            return self.url.rsplit("/", 1)[0].split("/", 1)[1]
        return ""

    @property
    def title(self):
        """
        Returns:
            title from meta
        """
        return self.meta.get("title", None)

    @property
    def tags(self):
        """
        Returns:
            tags from meta
        """
        return self.meta.get("tags", None)

    @property
    def date(self):
        """
        Returns:
            date from meta
        """
        meta = self.meta.get("date", None)

        if meta:
            return meta

        return datetime.now()

    @property
    def file_path(self):
        return "{0}/index.html".format(self.path)

    @property
    def images(self):
        """
        Returns:
            list of all user defined images
        """
        if self._images:
            return self._images

        parser = ImageParser()

        if hasattr(self, "rendered"):
            parser.feed(self.rendered)
        elif self.content is not None:
            parser.feed(self.content)
        else:
            return list()

        if len(parser.images) == 0:
            self._images = list()
        else:
            self._images = parser.images

        return self._images

    def get_url_value(self, key):
        """
        There are three possible scenarios where 'key' can be stored

          - part of the created timestamp
          - in the meta dictionary
          - as attribute of an Entity instance

        Arguments:
            key: key to lookup value for

        Returns:
            variable to be used in the URL for a given key
        """
        key = key[1:]  # remove %

        if hasattr(self.created, key):
            # return as string, not int
            return str(getattr(self.created, key))

        if key in self.meta:
            return self.meta[key]

        return getattr(self, key)



# -*- coding: utf-8 -*-

"""
    drupan.serve

    Serve a generated site locally.
"""

import os

from http.server import SimpleHTTPRequestHandler
from socketserver import TCPServer


def server(directory, port=9000):
    """
    Serve a directory over http.

    Arguments:
        directory: directory of the generated site
        port: port to run the httpd on
    """
    cwd = os.getcwd()
    os.chdir(directory)

    httpd = TCPServer(("", port), SimpleHTTPRequestHandler)

    try:
        print("server running on http://localhost:{0}".format(port))
        httpd.serve_forever()
    except KeyboardInterrupt:
        print("shutting down server\n")

    os.chdir(cwd)



# -*- coding: utf-8 -*-
"""
    drupan.plugins.blank

    Generate empty entities to be used for something that does not require
    its own content file.
"""

from drupan.entity import Entity


class Plugin(object):
    """create empty entities"""
    def __init__(self, site, config):
        """
        Arguments:
            site: instance of drupan.site.Site
            config: instance of drupan.config.Config
        """
        self.site = site
        self.config = config

        self.generate = config.get_option("blank", "generate")

        # 2.0 compatibility - 2.2 moved to a list where title and layout are
        # the same.
        if type(self.generate) == list:
            generate_dict = {}

            for item in self.generate:
                generate_dict[item] = item

            self.generate = generate_dict

    def run(self):
        """run the plugin"""
        for name in self.generate:
            new = Entity(self.config)
            new.meta["title"] = name
            new.meta["layout"] = self.generate[name]
            self.site.entities.append(new)



# -*- coding: utf-8 -*-
"""
    drupan.plugins.tags

    Create tag entities for all tagged objects.
"""

from drupan.entity import Entity


class Plugin(object):
    """convert entities content to markdown using markdown2"""
    def __init__(self, site, config):
        """
        Arguments:
            site: instance of drupan.site.Site
            config: instance of drupan.config.Config
        """
        self.site = site
        self.config = config

        self.tags = dict()

    def run(self):
        """run the plugin"""
        for entity in self.site.entities:
            self.generate_tags(entity)

        for tag in self.tags:
            self.site.entities.append(self.tags[tag])

        self.site.tags = self.tags

    def generate_tags(self, entity):
        """
        Generate entities for all tags in an entities meta information

        Arguments:
            instance of drupan.entity.Entity
        """
        if "tags" not in entity.meta:
            return

        for tag in entity.meta["tags"]:
            self.create_entity(tag)
            tag = tag.lower()
            self.tags[tag].entities.append(entity)

    def create_entity(self, tag):
        """
        Create a new entity for a tag if it does not exist

        Arguments:
            tag: tag name
        """
        lower = tag.lower()

        if lower in self.tags:
            return

        entity = Entity(self.config)
        entity.meta["title"] = tag
        entity.meta["layout"] = "tag"
        entity.entities = list()

        self.tags[lower] = entity



from distutils.core import setup

from drupan.version import __version__

try:
    import pypandoc
    long_description = pypandoc.convert("Readme.md", "rst")
except:
    long_description = "Readme not found"


setup(
    name="drupan",
    version=__version__,
    author="Timo Zimmermann",
    author_email="timo@screamingatmyscreen.com",
    packages=[
        "drupan",
        "drupan/plugins",
        "drupan/deployment",
        "drupan/inout"
    ],
    url="http://pypi.python.org/pypi/drupan/",
    license="BSD",
    description="Drupan is a flexible static site generator helping you \
    to create blogs, single page applications or traditional websites.",
    long_description=long_description,
    install_requires=[
        "pyyaml",
        "markdown2",
        "jinja2",
        "pygments",
    ],
    entry_points={
        "console_scripts": [
            "drupan = drupan.cmd:cmd"
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: BSD :: FreeBSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: News/Diary",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Software Development",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Code Generators",
    ],
)
